@@ -1,7 +1,21 @@
-# -*- test-case-name: buildbot.test.test_step -*-
+# This file is part of Buildbot.  Buildbot is free software: you can
+# redistribute it and/or modify it under the terms of the GNU General Public
+# License as published by the Free Software Foundation, version 2.
+#
+# This program is distributed in the hope that it will be useful, but WITHOUT
+# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
+# details.
+#
+# You should have received a copy of the GNU General Public License along with
+# this program; if not, write to the Free Software Foundation, Inc., 51
+# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+#
+# Copyright Buildbot Team Members
+
 
 from buildbot import util
-from buildbot.process.base import Build
+from buildbot.process.build import Build
 from buildbot.process.buildstep import BuildStep
 from buildbot.steps.source import CVS, SVN
 from buildbot.steps.shell import Configure, Compile, Test, PerlModuleTest
@@ -28,11 +42,12 @@ class ArgumentsInTheWrongPlace(Exception):
 class BuildFactory(util.ComparableMixin):
     """
     @cvar  buildClass: class to use when creating builds
-    @type  buildClass: L{buildbot.process.base.Build}
+    @type  buildClass: L{buildbot.process.build.Build}
     """
     buildClass = Build
     useProgress = 1
-    compare_attrs = ['buildClass', 'steps', 'useProgress']
+    workdir = "build"
+    compare_attrs = ['buildClass', 'steps', 'useProgress', 'workdir']
 
     def __init__(self, steps=None):
         if steps is None:
@@ -44,12 +59,15 @@ class BuildFactory(util.ComparableMixin):
             return step_or_factory.getStepFactory()
         return step_or_factory
 
-    def newBuild(self, request):
+    def newBuild(self, requests):
         """Create a new Build instance.
-        @param request: a L{base.BuildRequest} describing what is to be built
+
+        @param requests: a list of buildrequest dictionaries describing what is
+        to be built
         """
-        b = self.buildClass(request)
+        b = self.buildClass(requests)
         b.useProgress = self.useProgress
+        b.workdir = self.workdir
         b.setStepFactories(self.steps)
         return b
 
@@ -58,12 +76,16 @@ class BuildFactory(util.ComparableMixin):
             if kwargs:
                 raise ArgumentsInTheWrongPlace()
             s = step_or_factory.getStepFactory()
-        else:
+        elif type(step_or_factory) == type(BuildStep) and \
+                issubclass(step_or_factory, BuildStep):
             s = (step_or_factory, dict(kwargs))
+        else:
+            raise ValueError('%r is not a BuildStep nor BuildStep subclass' % step_or_factory)
         self.steps.append(s)
 
     def addSteps(self, steps):
-        self.steps.extend([ s.getStepFactory() for s in steps ])
+        for s in steps:
+            self.addStep(s)
 
 # BuildFactory subclasses for common build tools
 
