@@ -20,12 +20,35 @@
 # Also don't forget to mirror your changes on command-line options in manual
 # pages and texinfo documentation.
 
+import copy
 import os, sys, stat, re, time
-import traceback
 from twisted.python import usage, util, runtime
+from twisted.internet import defer
 
 from buildbot.interfaces import BuildbotNotRunningError
 
+def in_reactor(f):
+    """decorate a function by running it with maybeDeferred in a reactor"""
+    def wrap(*args, **kwargs):
+        from twisted.internet import reactor
+        result = [ ]
+        def async():
+            d = defer.maybeDeferred(f, *args, **kwargs)
+            def eb(f):
+                f.printTraceback()
+            d.addErrback(eb)
+            def do_stop(r):
+                result.append(r)
+                reactor.stop()
+            d.addBoth(do_stop)
+        reactor.callWhenRunning(async)
+        reactor.run()
+        return result[0]
+    wrap.__doc__ = f.__doc__
+    wrap.__name__ = f.__name__
+    wrap._orig = f # for tests
+    return wrap
+
 def isBuildmasterDir(dir):
     buildbot_tac = os.path.join(dir, "buildbot.tac")
     if not os.path.isfile(buildbot_tac):
@@ -52,15 +75,25 @@ class OptionsWithOptionsFile(usage.Options):
         # file, change the default in optParameters *before* calling through
         # to the parent constructor
 
-        if self.buildbotOptions:
-            optfile = loadOptionsFile()
-            for optfile_name, option_name in self.buildbotOptions:
-                for i in range(len(self.optParameters)):
-                    if self.optParameters[i][0] == option_name and optfile_name in optfile:
-                        self.optParameters[i][2] = optfile[optfile_name]
+        # Options uses reflect.accumulateClassList, so this *must* be
+        # a class attribute; however, we do not want to permanently change
+        # the class.  So we patch it temporarily and restore it after.
+        cls = self.__class__
+        if hasattr(cls, 'optParameters'):
+            old_optParameters = cls.optParameters
+            cls.optParameters = op = copy.deepcopy(cls.optParameters)
+            if self.buildbotOptions:
+                optfile = loadOptionsFile()
+                for optfile_name, option_name in self.buildbotOptions:
+                    for i in range(len(op)):
+                        if (op[i][0] == option_name and optfile_name in optfile):
+                            op[i] = list(op[i])
+                            op[i][2] = optfile[optfile_name]
         usage.Options.__init__(self, *args)
+        if hasattr(cls, 'optParameters'):
+            cls.optParameters = old_optParameters
 
-def loadOptionsFile(filename="options", here=None, home=None):
+def loadOptionsFile():
     """Find the .buildbot/FILENAME file. Crawl from the current directory up
     towards the root, and also look in ~/.buildbot . The first directory
     that's owned by the user and has the file we're looking for wins. Windows
@@ -71,18 +104,15 @@ def loadOptionsFile(filename="options", here=None, home=None):
              file was found, return an empty dict.
     """
 
-    if here is None:
-        here = os.getcwd()
-    here = os.path.abspath(here)
+    here = os.path.abspath(os.getcwd())
 
-    if home is None:
-        if runtime.platformType == 'win32':
-            # never trust env-vars, use the proper API
-            from win32com.shell import shellcon, shell            
-            appdata = shell.SHGetFolderPath(0, shellcon.CSIDL_APPDATA, 0, 0)
-            home = os.path.join(appdata, "buildbot")             
-        else:
-            home = os.path.expanduser("~/.buildbot")
+    if runtime.platformType == 'win32':
+        # never trust env-vars, use the proper API
+        from win32com.shell import shellcon, shell
+        appdata = shell.SHGetFolderPath(0, shellcon.CSIDL_APPDATA, 0, 0)
+        home = os.path.join(appdata, "buildbot")
+    else:
+        home = os.path.expanduser("~/.buildbot")
 
     searchpath = []
     toomany = 20
@@ -106,7 +136,7 @@ def loadOptionsFile(filename="options", here=None, home=None):
                 if os.stat(d)[stat.ST_UID] != os.getuid():
                     print "skipping %s because you don't own it" % d
                     continue # security, skip other people's directories
-            optfile = os.path.join(d, filename)
+            optfile = os.path.join(d, "options")
             if os.path.exists(optfile):
                 try:
                     f = open(optfile, "r")
@@ -128,7 +158,7 @@ class MakerBase(OptionsWithOptionsFile):
         ["quiet", "q", "Do not emit the commands being run"],
         ]
 
-    longdesc = """
+    usage.Options.longdesc = """
     Operates upon the specified <basedir> (or the current directory, if not
     specified).
     """
@@ -192,35 +222,6 @@ class Maker:
         if not self.quiet: print "mkdir", self.basedir
         os.mkdir(self.basedir)
 
-    def mkinfo(self):
-        path = os.path.join(self.basedir, "info")
-        if not os.path.exists(path):
-            if not self.quiet: print "mkdir", path
-            os.mkdir(path)
-        created = False
-        admin = os.path.join(path, "admin")
-        if not os.path.exists(admin):
-            if not self.quiet:
-                print "Creating info/admin, you need to edit it appropriately"
-            f = open(admin, "wt")
-            f.write("Your Name Here <admin@youraddress.invalid>\n")
-            f.close()
-            created = True
-        host = os.path.join(path, "host")
-        if not os.path.exists(host):
-            if not self.quiet:
-                print "Creating info/host, you need to edit it appropriately"
-            f = open(host, "wt")
-            f.write("Please put a description of this build host here\n")
-            f.close()
-            created = True
-        access_uri = os.path.join(path, "access_uri")
-        if not os.path.exists(access_uri):
-            if not self.quiet:
-                print "Not creating info/access_uri - add it if you wish"
-        if created and not self.quiet:
-            print "Please edit the files in %s appropriately." % path
-
     def chdir(self):
         if not self.quiet: print "chdir", self.basedir
         os.chdir(self.basedir)
@@ -296,16 +297,13 @@ class Maker:
             f.close()
 
     def create_db(self):
-        from buildbot.db import dbspec, exceptions
-        spec = dbspec.DBSpec.from_url(self.config["db"], self.basedir)
+        from buildbot.db import connector
+        from buildbot.master import BuildMaster
+        db = connector.DBConnector(BuildMaster(self.basedir),
+                self.config['db'], basedir=self.basedir)
         if not self.config['quiet']: print "creating database"
-
-        # upgrade from "nothing"
-        from buildbot.db.schema import manager
-        sm = manager.DBSchemaManager(spec, self.basedir)
-        if sm.get_db_version() != 0:
-            raise exceptions.DBAlreadyExistsError
-        sm.upgrade()
+        d = db.model.upgrade()
+        return d
 
     def populate_if_missing(self, target, source, overwrite=False):
         new_contents = open(source, "rt").read()
@@ -352,15 +350,17 @@ class Maker:
             self.populate_if_missing(os.path.join(webdir, target),
                                  source)
 
-    def check_master_cfg(self):
+    def check_master_cfg(self, expected_db_url=None):
+        """Check the buildmaster configuration, returning a deferred that
+        fires with an approprate exit status (so 0=success)."""
         from buildbot.master import BuildMaster
-        from twisted.python import log, failure
+        from twisted.python import log
 
         master_cfg = os.path.join(self.basedir, "master.cfg")
         if not os.path.exists(master_cfg):
             if not self.quiet:
                 print "No master.cfg found"
-            return 1
+            return defer.succeed(1)
 
         # side-effects of loading the config file:
 
@@ -378,31 +378,42 @@ class Maker:
             sys.path.insert(0, self.basedir)
 
         m = BuildMaster(self.basedir)
+
         # we need to route log.msg to stdout, so any problems can be seen
         # there. But if everything goes well, I'd rather not clutter stdout
         # with log messages. So instead we add a logObserver which gathers
         # messages and only displays them if something goes wrong.
         messages = []
         log.addObserver(messages.append)
-        try:
-            # this will raise an exception if there's something wrong with
-            # the config file. Note that this BuildMaster instance is never
-            # started, so it won't actually do anything with the
-            # configuration.
-            m.loadConfig(open(master_cfg, "r"), check_synchronously_only=True)
-        except:
-            f = failure.Failure()
+
+        # this will errback if there's something wrong with the config file.
+        # Note that this BuildMaster instance is never started, so it won't
+        # actually do anything with the configuration.
+        d = defer.maybeDeferred(lambda :
+            m.loadConfig(open(master_cfg, "r"), checkOnly=True))
+        def check_db_url(config):
+            if (expected_db_url and 
+                config.get('db_url', 'sqlite:///state.sqlite') != expected_db_url):
+                raise ValueError("c['db_url'] in the config file ('%s') does"
+                            " not match '%s'; please edit the configuration"
+                            " file before upgrading." %
+                                (config['db_url'], expected_db_url))
+        d.addCallback(check_db_url)
+        def cb(_):
+            return 0
+        def eb(f):
             if not self.quiet:
                 print
                 for m in messages:
                     print "".join(m['message'])
-                print f
+                f.printTraceback()
                 print
                 print "An error was detected in the master.cfg file."
                 print "Please correct the problem and run 'buildbot upgrade-master' again."
                 print
             return 1
-        return 0
+        d.addCallbacks(cb, eb)
+        return d
 
 DB_HELP = """
     The --db string is evaluated to build the DB object, which specifies
@@ -410,7 +421,6 @@ DB_HELP = """
     status information. The default (which creates an SQLite database in
     BASEDIR/state.sqlite) is equivalent to:
 
-      --db='DBSpec("sqlite3", basedir+"/state.sqlite"))'
       --db='sqlite:///state.sqlite'
 
     To use a remote MySQL database instead, use something like:
@@ -453,40 +463,53 @@ class UpgradeMasterOptions(MakerBase):
     pickle files.
     """
 
+@in_reactor
+@defer.deferredGenerator
 def upgradeMaster(config):
-    basedir = os.path.expanduser(config['basedir'])
     m = Maker(config)
+
+    if not config['quiet']: print "upgrading basedir"
+    basedir = os.path.expanduser(config['basedir'])
     # TODO: check Makefile
     # TODO: check TAC file
     # check web files: index.html, default.css, robots.txt
     m.upgrade_public_html({
-          'bg_gradient.jpg' : util.sibpath(__file__, "../status/web/files/bg_gradient.jpg"),
-          'default.css' : util.sibpath(__file__, "../status/web/files/default.css"),
-          'robots.txt' : util.sibpath(__file__, "../status/web/files/robots.txt"),
-          'favicon.ico' : util.sibpath(__file__, "../status/web/files/favicon.ico"),
-      })
+            'bg_gradient.jpg' : util.sibpath(__file__, "../status/web/files/bg_gradient.jpg"),
+            'default.css' : util.sibpath(__file__, "../status/web/files/default.css"),
+            'robots.txt' : util.sibpath(__file__, "../status/web/files/robots.txt"),
+            'favicon.ico' : util.sibpath(__file__, "../status/web/files/favicon.ico"),
+        })
     m.populate_if_missing(os.path.join(basedir, "master.cfg.sample"),
-                          util.sibpath(__file__, "sample.cfg"),
-                          overwrite=True)
+                            util.sibpath(__file__, "sample.cfg"),
+                            overwrite=True)
     # if index.html exists, use it to override the root page tempalte
     m.move_if_present(os.path.join(basedir, "public_html/index.html"),
-                      os.path.join(basedir, "templates/root.html"))
+                        os.path.join(basedir, "templates/root.html"))
 
-    from buildbot.db import dbspec
-    spec = dbspec.DBSpec.from_url(config["db"], basedir)
-    # TODO: check that TAC file specifies the right spec
+    if not config['quiet']: print "checking master.cfg"
+    wfd = defer.waitForDeferred(
+            m.check_master_cfg(expected_db_url=config['db']))
+    yield wfd
+    rc = wfd.getResult()
 
-    # upgrade the db
-    from buildbot.db.schema import manager
-    sm = manager.DBSchemaManager(spec, basedir)
-    sm.upgrade()
+    if rc == 0:
+        from buildbot.db import connector
+        from buildbot.master import BuildMaster
 
-    # check the configuration
-    rc = m.check_master_cfg()
-    if rc:
-        return rc
-    if not config['quiet']: print "upgrade complete"
-    return 0
+        if not config['quiet']: print "upgrading database"
+        db = connector.DBConnector(BuildMaster(config['basedir']),
+                            config['db'],
+                            basedir=config['basedir'])
+
+        wfd = defer.waitForDeferred(
+                db.model.upgrade())
+        yield wfd
+        wfd.getResult()
+
+        if not config['quiet']: print "upgrade complete"
+        yield 0
+    else:
+        yield rc
 
 
 class MasterOptions(MakerBase):
@@ -579,6 +602,7 @@ m.log_rotation.maxRotatedFiles = maxRotatedFiles
 
 """]
 
+@in_reactor
 def createMaster(config):
     m = Maker(config)
     m.mkdir()
@@ -599,9 +623,13 @@ def createMaster(config):
           'favicon.ico' : util.sibpath(__file__, "../status/web/files/favicon.ico"),
       })
     m.makefile()
-    m.create_db()
+    d = m.create_db()
 
-    if not m.quiet: print "buildmaster configured in %s" % m.basedir
+    def print_status(r):
+        if not m.quiet:
+            print "buildmaster configured in %s" % m.basedir
+    d.addCallback(print_status)
+    return d
 
 def stop(config, signame="TERM", wait=False):
     import signal
@@ -774,7 +802,9 @@ def statusgui(config):
     if master is None:
         raise usage.UsageError("master must be specified: on the command "
                                "line or in ~/.buildbot/options")
-    c = gtkPanes.GtkClient(master)
+    passwd = config.get('passwd')
+    username = config.get('username')
+    c = gtkPanes.GtkClient(master, username=username, passwd=passwd)
     c.run()
 
 class SendChangeOptions(OptionsWithOptionsFile):
@@ -790,6 +820,8 @@ class SendChangeOptions(OptionsWithOptionsFile):
         ("auth", "a", None, "Authentication token - username:password, or prompt for password"),
         ("who", "W", None, "Author of the commit"),
         ("repository", "R", '', "Repository specifier"),
+        ("vc", "s", None, "The VC system in use, one of: cvs, svn, darcs, hg, "
+                           "bzr, git, mtn, p4"),
         ("project", "P", '', "Project specifier"),
         ("branch", "b", None, "Branch specifier"),
         ("category", "C", None, "Category of repository"),
@@ -802,14 +834,18 @@ class SendChangeOptions(OptionsWithOptionsFile):
          "Read the log messages from this file (- for stdin)"),
         ("when", "w", None, "timestamp to use as the change time"),
         ("revlink", "l", '', "Revision link (revlink)"),
+        ("encoding", "e", 'utf8',
+            "Encoding of other parameters (default utf8)"),
         ]
 
     buildbotOptions = [
         [ 'master', 'master' ],
         [ 'who', 'who' ],
+        # deprecated in 0.8.3; remove in 0.8.5 (bug #1711)
         [ 'username', 'username' ],
         [ 'branch', 'branch' ],
         [ 'category', 'category' ],
+        [ 'vc', 'vc' ],
     ]
 
     def getSynopsis(self):
@@ -824,8 +860,9 @@ class SendChangeOptions(OptionsWithOptionsFile):
 def sendchange(config, runReactor=False):
     """Send a single change to the buildmaster's PBChangeSource. The
     connection will be drpoped as soon as the Change has been sent."""
-    from buildbot.clients.sendchange import Sender
+    from buildbot.clients import sendchange
 
+    encoding = config.get('encoding', 'utf8')
     who = config.get('who')
     if not who and config.get('username'):
         print "NOTE: --username/-u is deprecated: use --who/-W'"
@@ -837,6 +874,7 @@ def sendchange(config, runReactor=False):
     revision = config.get('revision')
     properties = config.get('properties', {})
     repository = config.get('repository', '')
+    vc = config.get('vc', None)
     project = config.get('project', '')
     revlink = config.get('revlink', '')
     if config.get('when'):
@@ -856,7 +894,11 @@ def sendchange(config, runReactor=False):
     if comments is None:
         comments = ""
 
-    files = config.get('files', [])
+    files = config.get('files', ())
+
+    vcs = ['cvs', 'svn', 'darcs', 'hg', 'bzr', 'git', 'mtn', 'p4', None]
+    assert vc in vcs, "vc must be 'cvs', 'svn', 'darcs', 'hg', 'bzr', " \
+        "'git', 'mtn', or 'p4'"
 
     # fix up the auth with a password if none was given
     if not auth:
@@ -870,18 +912,22 @@ def sendchange(config, runReactor=False):
     assert who, "you must provide a committer (--who)"
     assert master, "you must provide the master location"
 
-    s = Sender(master, auth)
+    s = sendchange.Sender(master, auth, encoding=encoding)
     d = s.send(branch, revision, comments, files, who=who, category=category, when=when,
-               properties=properties, repository=repository, project=project,
+               properties=properties, repository=repository, vc=vc, project=project,
                revlink=revlink)
+
     if runReactor:
+        from twisted.internet import reactor
         status = [True]
-        def failed(res):
+        def printSuccess(_):
+            print "change sent successfully"
+        def failed(f):
             status[0] = False
-            s.printFailure(res)
-        d.addCallbacks(s.printSuccess, failed)
-        d.addBoth(s.stop)
-        s.run()
+            print "change NOT sent - something went wrong: " + str(f)
+        d.addCallbacks(printSuccess, failed)
+        d.addBoth(lambda _ : reactor.stop())
+        reactor.run()
         return status[0]
     return d
 
@@ -892,6 +938,10 @@ class ForceOptions(OptionsWithOptionsFile):
         ["branch", None, None, "which branch to build"],
         ["revision", None, None, "which revision to build"],
         ["reason", None, None, "the reason for starting the build"],
+        ["props", None, None,
+         "A set of properties made available in the build environment, "
+         "format is --properties=prop1=value1,prop2=value2,.. "
+         "option can be specified multiple times."],
         ]
 
     def parseArgs(self, *args):
@@ -909,66 +959,92 @@ class ForceOptions(OptionsWithOptionsFile):
 class TryOptions(OptionsWithOptionsFile):
     optParameters = [
         ["connect", "c", None,
-         "how to reach the buildmaster, either 'ssh' or 'pb'"],
-        # for ssh, use --tryhost, --username, and --trydir
-        ["tryhost", None, None,
-         "the hostname (used by ssh) for the buildmaster"],
-        ["trydir", None, None,
-         "the directory (on the tryhost) where tryjobs are deposited"],
-        ["username", "u", None, "Username performing the trial build"],
+         "How to reach the buildmaster, either 'ssh' or 'pb'"],
+        # for ssh, use --host, --username, and --jobdir
+        ["host", None, None,
+         "Hostname (used by ssh) for the buildmaster"],
+        ["jobdir", None, None,
+         "Directory (on the buildmaster host) where try jobs are deposited"],
+        ["username", "u", None,
+         "Username performing the try build"],
         # for PB, use --master, --username, and --passwd
         ["master", "m", None,
          "Location of the buildmaster's PBListener (host:port)"],
-        ["passwd", None, None, "password for PB authentication"],
+        ["passwd", None, None,
+         "Password for PB authentication"],
+        ["who", "w", None,
+         "Who is responsible for the try build"],
+        ["comment", "C", None,
+         "A comment which can be used in notifications for this build"],
 
         ["diff", None, None,
-         "Filename of a patch to use instead of scanning a local tree. Use '-' for stdin."],
+         "Filename of a patch to use instead of scanning a local tree. "
+         "Use '-' for stdin."],
         ["patchlevel", "p", 0,
-         "Number of slashes to remove from patch pathnames, like the -p option to 'patch'"],
+         "Number of slashes to remove from patch pathnames, "
+         "like the -p option to 'patch'"],
 
         ["baserev", None, None,
          "Base revision to use instead of scanning a local tree."],
 
         ["vc", None, None,
-         "The VC system in use, one of: cvs,svn,bzr,darcs,p4"],
+         "The VC system in use, one of: bzr, cvs, darcs, git, hg, "
+         "mtn, p4, svn"],
         ["branch", None, None,
-         "The branch in use, for VC systems that can't figure it out"
-         " themselves"],
+         "The branch in use, for VC systems that can't figure it out "
+         "themselves"],
+        ["repository", None, None,
+         "Repository to use, instead of path to working directory."],
 
         ["builder", "b", None,
          "Run the trial build on this Builder. Can be used multiple times."],
         ["properties", None, None,
-         "A set of properties made available in the build environment, format:prop1=value1,prop2=value2..."],
+         "A set of properties made available in the build environment, "
+         "format is --properties=prop1=value1,prop2=value2,.. "
+         "option can be specified multiple times."],
 
-        ["try-topfile", None, None,
-         "Name of a file at the top of the tree, used to find the top. Only needed for SVN and CVS."],
-        ["try-topdir", None, None,
+        ["topfile", None, None,
+         "Name of a file at the top of the tree, used to find the top. "
+         "Only needed for SVN and CVS."],
+        ["topdir", None, None,
          "Path to the top of the working copy. Only needed for SVN and CVS."],
-
-        ]
+    ]
 
     optFlags = [
-        ["wait", None, "wait until the builds have finished"],
-        ["dryrun", 'n', "Gather info, but don't actually submit."],
-        ["get-builder-names", None, "Get the names of available builders. Doesn't submit anything. Only supported for 'pb' connections."],
-        ]
+        ["wait", None,
+         "wait until the builds have finished"],
+        ["dryrun", 'n',
+         "Gather info, but don't actually submit."],
+        ["get-builder-names", None,
+         "Get the names of available builders. Doesn't submit anything. "
+         "Only supported for 'pb' connections."],
+        ["quiet", "q",
+         "Don't print status of current builds while waiting."],
+    ]
 
-    # here it is, the definitive, quirky mapping of .buildbot/options names to
-    # command-line options.  Design by committee, anyone?
+    # Mapping of .buildbot/options names to command-line options
     buildbotOptions = [
         [ 'try_connect', 'connect' ],
         #[ 'try_builders', 'builders' ], <-- handled in postOptions
         [ 'try_vc', 'vc' ],
         [ 'try_branch', 'branch' ],
-        [ 'try_topdir', 'try-topdir' ],
-        [ 'try_topfile', 'try-topfile' ],
-        [ 'try_host', 'tryhost' ],
+        [ 'try_repository', 'repository' ],
+        [ 'try_topdir', 'topdir' ],
+        [ 'try_topfile', 'topfile' ],
+        [ 'try_host', 'host' ],
         [ 'try_username', 'username' ],
-        [ 'try_dir', 'trydir' ],
-        [ 'try_password', 'passwd' ],
+        [ 'try_jobdir', 'jobdir' ],
+        [ 'try_passwd', 'passwd' ],
         [ 'try_master', 'master' ],
+        [ 'try_who', 'who' ],
+        [ 'try_comment', 'comment' ],
         #[ 'try_wait', 'wait' ], <-- handled in postOptions
-        [ 'masterstatus', 'master' ],
+        #[ 'try_quiet', 'quiet' ], <-- handled in postOptions
+
+        # Deprecated command mappings from the quirky old days:
+        [ 'try_masterstatus', 'master' ],
+        [ 'try_dir', 'jobdir' ],
+        [ 'try_password', 'passwd' ],
     ]
 
     def __init__(self):
@@ -981,13 +1057,10 @@ class TryOptions(OptionsWithOptionsFile):
 
     def opt_properties(self, option):
         # We need to split the value of this option into a dictionary of properties
-        properties = {}
         propertylist = option.split(",")
         for i in range(0,len(propertylist)):
-            print propertylist[i]
-            splitproperty = propertylist[i].split("=")
-            properties[splitproperty[0]] = splitproperty[1]
-        self['properties'] = properties
+            splitproperty = propertylist[i].split("=", 1)
+            self['properties'][splitproperty[0]] = splitproperty[1]
 
     def opt_patchlevel(self, option):
         self['patchlevel'] = int(option)
@@ -1001,6 +1074,12 @@ class TryOptions(OptionsWithOptionsFile):
             self['builders'] = opts.get('try_builders', [])
         if opts.get('try_wait', False):
             self['wait'] = True
+        if opts.get('try_quiet', False):
+            self['quiet'] = True
+        # get the global 'masterstatus' option if it's set and no master
+        # was specified otherwise
+        if not self['master']:
+            self['master'] = opts.get('masterstatus', None)
 
 def doTry(config):
     from buildbot.clients import tryclient
@@ -1060,25 +1139,194 @@ class CheckConfigOptions(OptionsWithOptionsFile):
             self['configFile'] = 'master.cfg'
 
 
+@in_reactor
 def doCheckConfig(config):
+    from buildbot.scripts.checkconfig import ConfigLoader
     quiet = config.get('quiet')
     configFileName = config.get('configFile')
-    try:
-        from buildbot.scripts.checkconfig import ConfigLoader
-        if os.path.isdir(configFileName):
-            ConfigLoader(basedir=configFileName)
-        else:
-            ConfigLoader(configFileName=configFileName)
-    except:
+
+    if os.path.isdir(configFileName):
+        cl = ConfigLoader(basedir=configFileName)
+    else:
+        cl = ConfigLoader(configFileName=configFileName)
+
+    d = cl.load()
+
+    def cb(r):
         if not quiet:
-            # Print out the traceback in a nice format
-            t, v, tb = sys.exc_info()
-            traceback.print_exception(t, v, tb)
-        sys.exit(1)
+            print "Config file is good!"
+        return True
+    def eb(f):
+        if not quiet:
+            f.printTraceback()
+        return False
+    d.addCallbacks(cb, eb)
 
-    if not quiet:
-        print "Config file is good!"
+    return d
 
+class UserOptions(OptionsWithOptionsFile):
+    optParameters = [
+        ["master", "m", None,
+         "Location of the buildmaster's PBListener (host:port)"],
+        ["username", "u", None,
+         "Username for PB authentication"],
+        ["passwd", "p", None,
+         "Password for PB authentication"],
+        ["op", None, None,
+         "User management operation: add, remove, update, get"],
+        ["bb_username", None, None,
+         "Username to set for a given user. Only availabe on 'update', "
+         "and bb_password must be given as well."],
+        ["bb_password", None, None,
+         "Password to set for a given user. Only availabe on 'update', "
+         "and bb_username must be given as well."],
+        ["ids", None, None,
+         "User's identifiers, used to find users in 'remove' and 'get' "
+         "Can be specified multiple times (--ids=id1,id2,id3)"],
+        ["info", None, None,
+         "User information in the form: --info=type=value,type=value,.. "
+         "Used in 'add' and 'update', can be specified multiple times.  "
+         "Note that 'update' requires --info=id:type=value..."]
+    ]
+    buildbotOptions = [
+        [ 'master', 'master' ],
+        [ 'user_master', 'master' ],
+        [ 'user_username', 'username' ],
+        [ 'user_passwd', 'passwd' ],
+        ]
+
+    def __init__(self):
+        OptionsWithOptionsFile.__init__(self)
+        self['ids'] = []
+        self['info'] = []
+
+    def opt_ids(self, option):
+        id_list = option.split(",")
+        self['ids'].extend(id_list)
+
+    def opt_info(self, option):
+        # splits info into type/value dictionary, appends to info
+        info_list = option.split(",")
+        info_elem = {}
+
+        if len(info_list) == 1 and '=' not in info_list[0]:
+            info_elem["identifier"] = info_list[0]
+            self['info'].append(info_elem)
+        else:
+            for i in range(0, len(info_list)):
+                split_info = info_list[i].split("=", 1)
+
+                # pull identifier from update --info
+                if ":" in split_info[0]:
+                    split_id = split_info[0].split(":")
+                    info_elem["identifier"] = split_id[0]
+                    split_info[0] = split_id[1]
+
+                info_elem[split_info[0]] = split_info[1]
+            self['info'].append(info_elem)
+
+    def getSynopsis(self):
+        return "Usage:    buildbot user [options]"
+
+    longdesc = """
+    Currently implemented types for --info= are:\n
+    git, svn, hg, cvs, darcs, bzr, email
+    """
+
+def users_client(config, runReactor=False):
+    from buildbot.clients import usersclient
+    from buildbot.process.users import users    # for srcs, encrypt
+
+    # accepted attr_types by `buildbot user`, in addition to users.srcs
+    attr_types = ['identifier', 'email']
+
+    master = config.get('master')
+    assert master, "you must provide the master location"
+    try:
+        master, port = master.split(":")
+        port = int(port)
+    except:
+        raise AssertionError("master must have the form 'hostname:port'")
+
+    op = config.get('op')
+    assert op, "you must specify an operation: add, remove, update, get"
+    if op not in ['add', 'remove', 'update', 'get']:
+        raise AssertionError("bad op %r, use 'add', 'remove', 'update', "
+                             "or 'get'" % op)
+
+    username = config.get('username')
+    passwd = config.get('passwd')
+    assert username and passwd, "A username and password pair must be given"
+
+    bb_username = config.get('bb_username')
+    bb_password = config.get('bb_password')
+    if bb_username or bb_password:
+        if op != 'update':
+            raise AssertionError("bb_username and bb_password only work "
+                                 "with update")
+        if not bb_username or not bb_password:
+            raise AssertionError("Must specify both bb_username and "
+                                 "bb_password or neither.")
+
+        bb_password = users.encrypt(bb_password)
+
+    # check op and proper args
+    info = config.get('info')
+    ids = config.get('ids')
+
+    # check for erroneous args
+    if not info and not ids:
+        raise AssertionError("must specify either --ids or --info")
+    if info and ids:
+        raise AssertionError("cannot use both --ids and --info, use "
+                         "--ids for 'remove' and 'get', --info "
+                         "for 'add' and 'update'")
+
+    if op == 'add' or op == 'update':
+        if ids:
+            raise AssertionError("cannot use --ids with 'add' or 'update'")
+        if op == 'update':
+            for user in info:
+                if 'identifier' not in user:
+                    raise ValueError("no ids found in update info, use: "
+                                     "--info=id:type=value,type=value,..")
+        if op == 'add':
+            for user in info:
+                if 'identifier' in user:
+                    raise ValueError("id found in add info, use: "
+                                     "--info=type=value,type=value,..")
+    if op == 'remove' or op == 'get':
+        if info:
+            raise AssertionError("cannot use --info with 'remove' or 'get'")
+
+    # find identifier if op == add
+    if info:
+        # check for valid types
+        for user in info:
+            for attr_type in user:
+                if attr_type not in users.srcs + attr_types:
+                    raise ValueError("Type not a valid attr_type, must be in: "
+                                     "%r" % (users.srcs + attr_types))
+
+            if op == 'add':
+                user['identifier'] = user.values()[0]
+
+    uc = usersclient.UsersClient(master, username, passwd, port)
+    d = uc.send(op, bb_username, bb_password, ids, info)
+
+    if runReactor:
+        from twisted.internet import reactor
+        status = [True]
+        def printSuccess(res):
+            print res
+        def failed(f):
+            status[0] = False
+            print "user op NOT sent - something went wrong: " + str(f)
+        d.addCallbacks(printSuccess, failed)
+        d.addBoth(lambda _ : reactor.stop())
+        reactor.run()
+        return status[0]
+    return d
 
 class Options(usage.Options):
     synopsis = "Usage:    buildbot <command> [command options]"
@@ -1119,6 +1367,9 @@ class Options(usage.Options):
         ['checkconfig', None, CheckConfigOptions,
          "test the validity of a master.cfg config file"],
 
+        ['user', None, UserOptions,
+         "Manage users in buildbot's database"]
+
         # TODO: 'watch'
         ]
 
@@ -1189,6 +1440,10 @@ def run():
     elif command == "tryserver":
         doTryServer(so)
     elif command == "checkconfig":
-        doCheckConfig(so)
+        if not doCheckConfig(so):
+            sys.exit(1)
+    elif command == "user":
+        if not users_client(so, True):
+            sys.exit(1)
     sys.exit(0)
 
