@@ -15,10 +15,11 @@
 
 
 import re
-
+import types
 from email.Message import Message
 from email.Utils import formatdate
 from email.MIMEText import MIMEText
+from email.MIMENonMultipart import MIMENonMultipart
 from email.MIMEMultipart import MIMEMultipart
 from StringIO import StringIO
 import urllib
@@ -36,8 +37,9 @@ except ImportError:
     have_ssl = False
 
 from buildbot import interfaces, util
+from buildbot.process.users import users
 from buildbot.status import base
-from buildbot.status.builder import FAILURE, SUCCESS, Results
+from buildbot.status.results import FAILURE, SUCCESS, Results
 
 VALID_EMAIL = re.compile("[a-zA-Z0-9\.\_\%\-\+]+@[a-zA-Z0-9\.\_\%\-]+.[a-zA-Z]{2,6}")
 
@@ -80,7 +82,7 @@ def defaultMessage(mode, name, build, results, master_status):
     if ss and ss.project:
         project = ss.project
     else:
-        project = master_status.getProjectName()
+        project = master_status.getTitle()
     text += " on builder %s while building %s.\n" % (name, project)
     if master_status.getURLForThing(build):
         text += "Full details are available at:\n %s\n" % master_status.getURLForThing(build)
@@ -153,9 +155,11 @@ class MailNotifier(base.StatusReceiverMultiService):
                      "subject", "sendToInterestedUsers", "customMesg",
                      "messageFormatter", "extraHeaders"]
 
+    possible_modes = ('all', 'failing', 'problem', 'change', 'passing', 'warnings')
+
     def __init__(self, fromaddr, mode="all", categories=None, builders=None,
-                 addLogs=False, relayhost="localhost",
-                 subject="buildbot %(result)s in %(projectName)s on %(builder)s",
+                 addLogs=False, relayhost="localhost", buildSetSummary=False,
+                 subject="buildbot %(result)s in %(title)s on %(builder)s",
                  lookup=None, extraRecipients=[],
                  sendToInterestedUsers=True, customMesg=None,
                  messageFormatter=defaultMessage, extraHeaders=None,
@@ -186,7 +190,7 @@ class MailNotifier(base.StatusReceiverMultiService):
                         builder which provoked the message.
 
         @type  mode: string (defaults to all)
-        @param mode: one of:
+        @param mode: one of MailNotifer.possible_modes:
                      - 'all': send mail about all builds, passing and failing
                      - 'failing': only send mail about builds which fail
                      - 'warnings': send mail if builds contain warnings or fail 
@@ -219,7 +223,12 @@ class MailNotifier(base.StatusReceiverMultiService):
         @type  relayhost: string
         @param relayhost: the host to which the outbound SMTP connection
                           should be made. Defaults to 'localhost'
-
+                          
+        @type  buildSetSummary: boolean
+        @param buildSetSummary: if True, this notifier will only send a summary
+                                email when a buildset containing any of its
+                                watched builds completes
+                                
         @type  lookup:    implementor of {IEmailLookup}
         @param lookup:    object which provides IEmailLookup, which is
                           responsible for mapping User names for Interested
@@ -280,7 +289,7 @@ class MailNotifier(base.StatusReceiverMultiService):
         self.extraRecipients = extraRecipients
         self.sendToInterestedUsers = sendToInterestedUsers
         self.fromaddr = fromaddr
-        assert mode in ('all', 'failing', 'problem', 'change', 'passing', 'warnings')
+        assert mode in self.possible_modes
         self.mode = mode
         self.categories = categories
         self.builders = builders
@@ -302,6 +311,8 @@ class MailNotifier(base.StatusReceiverMultiService):
         self.smtpUser = smtpUser
         self.smtpPassword = smtpPassword
         self.smtpPort = smtpPort
+        self.buildSetSummary = buildSetSummary
+        self.buildSetSubscription = None
         self.watched = []
         self.master_status = None
 
@@ -323,6 +334,22 @@ class MailNotifier(base.StatusReceiverMultiService):
     def setup(self):
         self.master_status = self.parent.getStatus()
         self.master_status.subscribe(self)
+        
+            
+    def startService(self):
+        if self.buildSetSummary:
+            self.buildSetSubscription = \
+            self.parent.subscribeToBuildsetCompletions(self.buildsetFinished)
+ 
+        base.StatusReceiverMultiService.startService(self)
+        
+   
+    def stopService(self):
+        if self.buildSetSubscription is not None:
+            self.buildSetSubscription.unsubscribe()
+            self.buildSetSubscription = None
+            
+        return base.StatusReceiverMultiService.stopService(self)
 
     def disownServiceParent(self):
         self.master_status.unsubscribe(self)
@@ -345,38 +372,79 @@ class MailNotifier(base.StatusReceiverMultiService):
         pass
     def buildStarted(self, name, build):
         pass
-    def buildFinished(self, name, build, results):
+    def isMailNeeded(self, build, results):
         # here is where we actually do something.
         builder = build.getBuilder()
-        if self.builders is not None and name not in self.builders:
-            return # ignore this build
+        if self.builders is not None and builder.name not in self.builders:
+            return False # ignore this build
         if self.categories is not None and \
                builder.category not in self.categories:
-            return # ignore this build
+            return False # ignore this build
 
         if self.mode == "warnings" and results == SUCCESS:
-            return
+            return False
         if self.mode == "failing" and results != FAILURE:
-            return
+            return False
         if self.mode == "passing" and results != SUCCESS:
-            return
+            return False
         if self.mode == "problem":
             if results != FAILURE:
-                return
+                return False
             prev = build.getPreviousBuild()
             if prev and prev.getResults() == FAILURE:
-                return
+                return False
         if self.mode == "change":
             prev = build.getPreviousBuild()
             if not prev or prev.getResults() == results:
-                return
-        # for testing purposes, buildMessage returns a Deferred that fires
-        # when the mail has been sent. To help unit tests, we return that
-        # Deferred here even though the normal IStatusReceiver.buildFinished
-        # signature doesn't do anything with it. If that changes (if
-        # .buildFinished's return value becomes significant), we need to
-        # rearrange this.
-        return self.buildMessage(name, build, results)
+                return False
+        
+        return True
+
+    def buildFinished(self, name, build, results):
+        if ( not self.buildSetSummary and
+             self.isMailNeeded(build, results) ):
+            # for testing purposes, buildMessage returns a Deferred that fires
+            # when the mail has been sent. To help unit tests, we return that
+            # Deferred here even though the normal IStatusReceiver.buildFinished
+            # signature doesn't do anything with it. If that changes (if
+            # .buildFinished's return value becomes significant), we need to
+            # rearrange this.
+            return self.buildMessage(name, [build], results)
+        return None
+    
+    def _gotBuilds(self, res, builddicts, buildset, builders):
+        builds = []
+        for (builddictlist, builder) in zip(builddicts, builders):
+                for builddict in builddictlist:
+                    build = builder.getBuild(builddict['number'])
+                    if self.isMailNeeded(build, build.results):
+                        builds.append(build)
+
+        self.buildMessage("Buildset Complete: " + buildset['reason'], builds,
+                          buildset['results'])
+        
+    def _gotBuildRequests(self, breqs, buildset):
+        builddicts = []
+        builders =[]
+        dl = []
+        for breq in breqs:
+            buildername = breq['buildername']
+            builders.append(self.master_status.getBuilder(buildername))
+            d = self.parent.db.builds.getBuildsForRequest(breq['brid'])
+            d.addCallback(builddicts.append)
+            dl.append(d)
+        d = defer.DeferredList(dl)
+        d.addCallback(self._gotBuilds, builddicts, buildset, builders)
+
+    def _gotBuildSet(self, buildset, bsid):
+        d = self.parent.db.buildrequests.getBuildRequests(bsid=bsid)
+        d.addCallback(self._gotBuildRequests, buildset)
+        
+    def buildsetFinished(self, bsid, result):
+        d = self.parent.db.buildsets.getBuildset(bsid=bsid)
+        d.addCallback(self._gotBuildSet, bsid)
+            
+        return d
 
     def getCustomMesgData(self, mode, name, build, results, master_status):
         #
@@ -390,12 +458,14 @@ class MailNotifier(base.StatusReceiverMultiService):
             logStatus, dummy = logStep.getResults()
             logName = logf.getName()
             logs.append(('%s.%s' % (stepName, logName),
-                         '%s/steps/%s/logs/%s' % (master_status.getURLForThing(build), stepName, logName),
+                         '%s/steps/%s/logs/%s' % (
+                             master_status.getURLForThing(build),
+                             stepName, logName),
                          logf.getText().splitlines(),
                          logStatus))
 
         attrs = {'builderName': name,
-                 'projectName': master_status.getProjectName(),
+                 'title': master_status.getTitle(),
                  'mode': mode,
                  'result': Results[results],
                  'buildURL': master_status.getURLForThing(build),
@@ -408,6 +478,7 @@ class MailNotifier(base.StatusReceiverMultiService):
                  'branch': "",
                  'revision': "",
                  'patch': "",
+                 'patch_info': "",
                  'changes': [],
                  'logs': logs}
 
@@ -416,26 +487,52 @@ class MailNotifier(base.StatusReceiverMultiService):
             attrs['branch'] = ss.branch
             attrs['revision'] = ss.revision
             attrs['patch'] = ss.patch
+            attrs['patch_info'] = ss.patch_info
             attrs['changes'] = ss.changes[:]
 
         return attrs
 
-    def createEmail(self, msgdict, builderName, projectName, results, build,
-                    patch=None, logs=None):
+    def patch_to_attachment(self, patch, index):
+        # patches don't come with an encoding.  If the patch is valid utf-8,
+        # we'll attach it as MIMEText; otherwise, it gets attached as a binary
+        # file.  This will suit the vast majority of cases, since utf8 is by
+        # far the most common encoding.
+        if type(patch[1]) != types.UnicodeType:
+            try:
+                    unicode = patch[1].decode('utf8')
+            except UnicodeDecodeError:
+                unicode = None
+        else:
+            unicode = patch[1]
+
+        if unicode:
+            a = MIMEText(unicode.encode(ENCODING), _charset=ENCODING)
+        else:
+            # MIMEApplication is not present in Python-2.4 :(
+            a = MIMENonMultipart('application', 'octet-stream')
+            a.set_payload(patch[1])
+        a.add_header('Content-Disposition', "attachment",
+                    filename="source patch " + str(index) )
+        return a
+
+    def createEmail(self, msgdict, builderName, title, results, builds=None,
+                    patches=None, logs=None):
         text = msgdict['body'].encode(ENCODING)
         type = msgdict['type']
         if 'subject' in msgdict:
             subject = msgdict['subject'].encode(ENCODING)
         else:
             subject = self.subject % { 'result': Results[results],
-                                       'projectName': projectName,
+                                       'projectName': title,
+                                       'title': title,
                                        'builder': builderName,
                                        }
 
 
-        assert type in ('plain', 'html'), "'%s' message type must be 'plain' or 'html'." % type
+        assert type in ('plain', 'html'), \
+            "'%s' message type must be 'plain' or 'html'." % type
 
-        if patch or logs:
+        if patches or logs:
             m = MIMEMultipart()
             m.attach(MIMEText(text, type, ENCODING))
         else:
@@ -448,69 +545,134 @@ class MailNotifier(base.StatusReceiverMultiService):
         m['From'] = self.fromaddr
         # m['To'] is added later
 
-        if patch:
-            a = MIMEText(patch[1].encode(ENCODING), _charset=ENCODING)
-            a.add_header('Content-Disposition', "attachment",
-                         filename="source patch")
-            m.attach(a)
+        if patches:
+            for (i, patch) in enumerate(patches):
+                a = self.patch_to_attachment(patch, i)
+                m.attach(a)
         if logs:
             for log in logs:
                 name = "%s.%s" % (log.getStep().getName(),
                                   log.getName())
-                if self._shouldAttachLog(log.getName()) or self._shouldAttachLog(name):
+                if ( self._shouldAttachLog(log.getName()) or
+                     self._shouldAttachLog(name) ):
                     a = MIMEText(log.getText().encode(ENCODING), 
                                  _charset=ENCODING)
                     a.add_header('Content-Disposition', "attachment",
                                  filename=name)
                     m.attach(a)
 
+        #@todo: is there a better way to do this?
         # Add any extra headers that were requested, doing WithProperties
-        # interpolation if necessary
+        # interpolation if only one build was given
         if self.extraHeaders:
-            properties = build.getProperties()
             for k,v in self.extraHeaders.items():
-                k = properties.render(k)
+                if len(builds) == 1:
+                    k = interfaces.IProperties(builds[0]).render(k)
                 if k in m:
-                    twlog.msg("Warning: Got header " + k + " in self.extraHeaders "
-                          "but it already exists in the Message - "
-                          "not adding it.")
-                    continue
-                m[k] = properties.render(v)
-
+                    twlog.msg("Warning: Got header " + k +
+                      " in self.extraHeaders "
+                      "but it already exists in the Message - "
+                      "not adding it.")
+                if len(builds) == 1:
+                    m[k] = interfaces.IProperties(builds[0]).render(v)
+                else:
+                    m[k] = v
+    
         return m
-
-    def buildMessage(self, name, build, results):
+    
+    def buildMessageDict(self, name, build, results):
         if self.customMesg:
             # the customMesg stuff can be *huge*, so we prefer not to load it
-            attrs = self.getCustomMesgData(self.mode, name, build, results, self.master_status)
+            attrs = self.getCustomMesgData(self.mode, name, build, results,
+                                           self.master_status)
             text, type = self.customMesg(attrs)
             msgdict = { 'body' : text, 'type' : type }
         else:
-            msgdict = self.messageFormatter(self.mode, name, build, results, self.master_status)
+            msgdict = self.messageFormatter(self.mode, name, build, results,
+                                            self.master_status)
+        
+        return msgdict
 
-        patch = None
-        ss = build.getSourceStamp()
-        if ss and ss.patch and self.addPatch:
-            patch == ss.patch
-        logs = None
-        if self.addLogs:
-            logs = build.getLogs()
-            twlog.err("LOG: %s" % str(logs))
-        m = self.createEmail(msgdict, name, self.master_status.getProjectName(),
-                             results, build, patch, logs)
+
+    def buildMessage(self, name, builds, results):
+        patches = []
+        logs = []
+        msgdict = {"body":""}
+        
+        for build in builds:
+            ss = build.getSourceStamp()
+            if ss and ss.patch and self.addPatch:
+                patches.append(ss.patch)
+            if self.addLogs:
+                logs.extend(build.getLogs())
+            
+            tmp = self.buildMessageDict(name=build.getBuilder().name,
+                                        build=build, results=build.results)
+            msgdict['body'] += tmp['body']
+            msgdict['body'] += '\n\n'
+            msgdict['type'] = tmp['type']
+            if "subject" in tmp:
+                msgdict['subject'] = tmp['subject']
+
+        m = self.createEmail(msgdict, name, self.master_status.getTitle(),
+                             results, builds, patches, logs)
 
         # now, who is this message going to?
-        dl = []
-        recipients = []
-        if self.sendToInterestedUsers and self.lookup:
-            for u in build.getInterestedUsers():
-                d = defer.maybeDeferred(self.lookup.getAddress, u)
-                d.addCallback(recipients.append)
-                dl.append(d)
-        d = defer.DeferredList(dl)
-        d.addCallback(self._gotRecipients, recipients, m)
+        self.dl = []
+        self.recipients = []
+        if self.sendToInterestedUsers:
+            for build in builds:
+                d = defer.succeed(build)
+                if self.lookup:
+                    d.addCallback(self.useLookup)
+                else:
+                    d.addCallback(self.useUsers)
+        else:
+            d = defer.DeferredList(self.dl)
+        d.addCallback(self._gotRecipients, self.recipients, m)
         return d
 
+    def useLookup(self, build):
+        for u in build.getInterestedUsers():
+            d = defer.maybeDeferred(self.lookup.getAddress, u)
+            d.addCallback(self.recipients.append)
+            self.dl.append(d)
+        return defer.DeferredList(self.dl)
+
+    def useUsers(self, build):
+        self.contacts = []
+        ss = build.getSourceStamp()
+        for change in ss.changes:
+            d = self.parent.db.changes.getChangeUids(change.number)
+            def getContacts(uids):
+                def uidContactPair(contact, uid):
+                    return (contact, uid)
+                d = defer.succeed(None)
+                for uid in uids:
+                    d.addCallback(lambda _ :
+                                     users.getUserContact(self.parent,
+                                                          contact_type='email',
+                                                          uid=uid))
+                    d.addCallback(lambda contact: uidContactPair(contact, uid))
+                    d.addCallback(self.contacts.append)
+                return d
+            d.addCallback(getContacts)
+            def logNoMatch(_):
+                for pair in self.contacts:
+                    contact, uid = pair
+                    if contact is None:
+                        twlog.msg("Unable to find email for uid: %r" % uid)
+                return [pair[0] for pair in self.contacts]
+            d.addCallback(logNoMatch)
+            d.addCallback(self.recipients.extend)
+            def addOwners(_):
+                owners = [e for e in build.getInterestedUsers()
+                          if e not in build.getResponsibleUsers()]
+                self.recipients.extend(owners)
+            d.addCallback(addOwners)
+            self.dl.append(d)
+        return defer.DeferredList(self.dl)
+
     def _shouldAttachLog(self, logname):
         if type(self.addLogs) is bool:
             return self.addLogs
