@@ -1,13 +1,52 @@
+# This file is part of Buildbot.  Buildbot is free software: you can
+# redistribute it and/or modify it under the terms of the GNU General Public
+# License as published by the Free Software Foundation, version 2.
+#
+# This program is distributed in the hope that it will be useful, but WITHOUT
+# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
+# details.
+#
+# You should have received a copy of the GNU General Public License along with
+# this program; if not, write to the Free Software Foundation, Inc., 51
+# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+#
+# Copyright Buildbot Team Members
+
 
 import time, urllib
 from twisted.web import html
 from twisted.web.util import Redirect
 from twisted.web.error import NoResource
+from twisted.internet import defer
 
 from buildbot.status.web.base import HtmlResource, abbreviate_age, \
-    BuildLineMixin, path_to_slave, path_to_authfail
+    BuildLineMixin, ActionResource, path_to_slave, path_to_authfail
 from buildbot import util
 
+class ShutdownActionResource(ActionResource):
+
+    def __init__(self, slave):
+        self.slave = slave
+        self.action = "gracefulShutdown"
+
+    @defer.deferredGenerator
+    def performAction(self, request):
+        d = self.getAuthz(request).actionAllowed(self.action,
+                                                 request,
+                                                 self.slave)
+        wfd = defer.waitForDeferred(d)
+        yield wfd
+        res = wfd.getResult()
+
+        url = None
+        if res:
+            self.slave.setGraceful(True)
+            url = path_to_slave(request, self.slave)
+        else:
+            url = path_to_authfail(request)
+        yield url
+
 # /buildslaves/$slavename
 class OneBuildSlaveResource(HtmlResource, BuildLineMixin):
     addSlash = False
@@ -15,17 +54,14 @@ class OneBuildSlaveResource(HtmlResource, BuildLineMixin):
         HtmlResource.__init__(self)
         self.slavename = slavename
 
-    def getTitle(self, req):
+    def getPageTitle(self, req):
         return "Buildbot: %s" % self.slavename
 
     def getChild(self, path, req):
         s = self.getStatus(req)
         slave = s.getSlave(self.slavename)
         if path == "shutdown":
-            if self.getAuthz(req).actionAllowed("gracefulShutdown", req, slave):
-                slave.setGraceful(True)
-            else:
-                return Redirect(path_to_authfail(req))
+            return ShutdownActionResource(slave)
         return Redirect(path_to_slave(req, slave))
 
     def content(self, request, ctx):        
@@ -74,7 +110,7 @@ class OneBuildSlaveResource(HtmlResource, BuildLineMixin):
                         access_uri = slave.getAccessURI()),
                         admin = unicode(slave.getAdmin() or '', 'utf-8'),
                         host = unicode(slave.getHost() or '', 'utf-8'),
-                        version = slave.getVersion(),
+                        slave_version = slave.getVersion(),
                         show_builder_column = True,
                         connect_count = connect_count)
         template = request.site.buildbot_service.templates.get_template("buildslave.html")
@@ -83,7 +119,7 @@ class OneBuildSlaveResource(HtmlResource, BuildLineMixin):
 
 # /buildslaves
 class BuildSlavesResource(HtmlResource):
-    title = "BuildSlaves"
+    pageTitle = "BuildSlaves"
     addSlash = True
 
     def content(self, request, ctx):
@@ -121,13 +157,12 @@ class BuildSlavesResource(HtmlResource):
             info['connected'] = slave.isConnected()
             info['connectCount'] = slave.getConnectCount()
             
-            if slave.isConnected():
-                info['admin'] = unicode(slave.getAdmin() or '', 'utf-8')
-                last = slave.lastMessageReceived()
-                if last:
-                    info['last_heard_from_age'] = abbreviate_age(time.time() - last)
-                    info['last_heard_from_time'] = time.strftime("%Y-%b-%d %H:%M:%S",
-                                                                time.localtime(last))
+            info['admin'] = unicode(slave.getAdmin() or '', 'utf-8')
+            last = slave.lastMessageReceived()
+            if last:
+                info['last_heard_from_age'] = abbreviate_age(time.time() - last)
+                info['last_heard_from_time'] = time.strftime("%Y-%b-%d %H:%M:%S",
+                                                            time.localtime(last))
 
         template = request.site.buildbot_service.templates.get_template("buildslaves.html")
         data = template.render(**ctx)
